@@ -1,8 +1,8 @@
+import datetime
 import logging
 import os
 from collections import namedtuple
-from functools import partial
-from enum import Enum
+from decimal import Decimal
 from os import sys
 from winreg import HKEY_LOCAL_MACHINE, OpenKey, QueryValueEx
 
@@ -26,22 +26,21 @@ class InvalidFilter(Exception):
 
 class VismaAPI:
     """
-    Call VismaAPI to open connection with a database (visma company folder)
+    Opens a connection to Visma database
 
     Example:
-        VismaAPI(common_path="Z:\\Gemensamma filer", company_path="Z:\\Företag\\FTG9")
+        my_company = VismaAPI(common_path="Z:\\Gemensamma filer", company_path="Z:\\Företag\\FTG9")
+
+    The class expects you to either provide username and password as keyword arguments upon instantiation,
+    or supply them through visma_username and visma_password environment variables
 
     The API may be accessed directly with the .api attribute,
     this property exposes the C# DLL and gives you access to all functionality
     defined in AdkNet4Wrapper.dll and Adk.h
-
-    Example:
-        VismaAPI().api.AdkSetFilter() # AdkSetFilter is a method of the dll.
     """
 
-    _API = None
-
     def __init__(self, *args, **kwargs):
+        self._api = None
         self.adk = str()
         self.common_path = str()
         self.company_path = str()
@@ -51,12 +50,27 @@ class VismaAPI:
             self.common_path = kwargs.pop("common_path", self.common_path)
             self.company_path = kwargs.pop("company_path", self.company_path)
 
-        # credentials for AdkOpen
-        login = self.get_login_credentials()
-        self.username = login.username
-        self.password = login.password
+        if "username" and "password" in kwargs:
+            self.username = kwargs.pop("username")
+            self.password = kwargs.pop("password")
+        else:
+            login = self.get_login_credentials()
+            self.username = login.username
+            self.password = login.password
 
-        self.__class__._API = self.api
+        self.available_fields = {
+            self.field_without_db_prefix(field).lower(): field
+            for field in self.db_fields()
+        }
+
+    def __getattr__(self, name):
+
+        if name in self.available_fields:
+            return type(
+                name.title(), (_DBField,), {"DB_NAME": self.available_fields[name]}
+            )(api=self.api)
+
+        raise AttributeError
 
     def __del__(self):
         self.api.AdkClose()
@@ -71,8 +85,8 @@ class VismaAPI:
 
     @property
     def api(self):
-        if self._API:
-            return self._API
+        if self._api:
+            return self._api
 
         error = Api.AdkOpen2(
             self.common_path, self.company_path, self.username, self.password
@@ -82,6 +96,7 @@ class VismaAPI:
             logging.error(error_message)
             sys.exit(1)
 
+        self._api = Api
         return Api
 
     @staticmethod
@@ -104,26 +119,6 @@ class VismaAPI:
         fields = [field for field in Api.__dict__ if field.startswith("ADK_DB")]
         return fields
 
-    @staticmethod
-    def fields_for_db_field(db_field):
-        """
-        Example valid db_field is ADK_DB_SUPPLIER ( last word being the keyword )
-
-        Fields of type ADK_DB_SUPPLIER are defined as ADK_SUPPLIER_FIELDNAME
-
-        Function returns a list of names of associated fields of type db_field
-        """
-
-        if db_field not in VismaAPI.db_fields():
-            logging.error(f"{db_field} is not listed as a db field in Adk.h")
-
-        fields = [
-            field
-            for field in Api.__dict__
-            if field.startswith("ADK_" + VismaAPI.field_without_db_prefix(db_field))
-        ]
-        return fields
-
     @staticmethod
     def field_without_db_prefix(db_field):
         return db_field.replace("ADK_DB_", "")
@@ -133,9 +128,10 @@ class _DBField:
 
     DB_NAME = None
 
-    def __init__(self):
-        self.api = VismaAPI._API
+    def __init__(self, api):
+        self.api = api
         self.pdata = _Pdata(
+            self.api,
             self.__class__.DB_NAME,
             self.api.AdkCreateData(getattr(self.api, self.DB_NAME)),
         )
@@ -148,12 +144,21 @@ class _DBField:
         Example:
 
             Supplier().filter(A="a", B="b") # Only applies filtering on B
-            # B must be a valid field on ADK_DB_SUPPLIER
-            
+            # B must be a valid field of ADK_DB_SUPPLIER
+
         """
         for field, filter_term in kwargs.items():
-            if field not in VismaAPI.fields_for_db_field(self.__class__.DB_NAME):
-                raise Exception("Field not found.")
+            field = field.upper()  # Fields in Visma are all uppercased
+            try:
+                self.api.AdkGetType(
+                    self.pdata.data,
+                    getattr(self.api, field),
+                    self.api.ADK_FIELD_TYPE.eUnused,
+                )
+            except AttributeError:
+                raise AttributeError(
+                    f"{field} is not a valid field of {self.__class__.DB_NAME}"
+                )
 
             error = self.api.AdkSetFilter(
                 self.pdata.data, getattr(self.api, field), filter_term, 0
@@ -161,12 +166,15 @@ class _DBField:
             if error.lRc != self.api.ADKE_OK:
                 raise InvalidFilter
 
+    def new(self):
+        return self.pdata
+
     def get(self, **kwargs):
         """
         Returns a single object, or returns an exception.
         """
         self.set_filter(**kwargs)
-        error = self.api.AdkFirstEx(self.pdata.data, False)
+        error = self.api.AdkFirstEx(self.pdata.data, True)
         if error.lRc != self.api.ADKE_OK:
             error_message = Api.AdkGetErrorText(
                 error, self.api.ADK_ERROR_TEXT_TYPE.elRc
@@ -185,12 +193,8 @@ class _DBField:
         except Exception:
             return
 
-        # while (
-        #    error := self.api.AdkNextEx(self.pdata.data, False).lRc == self.api.ADKE_OK
-        # ) :
-
         while True:
-            error = self.api.AdkNextEx(self.pdata.data, False).lRc
+            error = self.api.AdkNextEx(self.pdata.data, True).lRc
             if error != self.api.ADKE_OK:
                 break
 
@@ -205,9 +209,9 @@ class _Pdata(object):
     Access and set fields like normal instance attributes
 
     Example:
-    
+
         # hello is an instance of Pdata which data is of type ADK_DB_SUPPLIER
-        hello = Supplier().get(ADK_SUPPLIER_NAME="hello")
+        hello = visma.supplier.get(ADK_SUPPLIER_NAME="hello")
 
         # Access a field on hello
         hello.adk_supplier_name
@@ -215,37 +219,39 @@ class _Pdata(object):
         # Set a field on hello
         hello.adk_supplier_name = "hello1"
         hello.save()
-
-        
     """
 
-    def __init__(self, db_name, pdata):
-        object.__setattr__(self, "api", VismaAPI._API)
+    def __init__(self, api, db_name, pdata):
+        object.__setattr__(self, "api", api)
         object.__setattr__(self, "db_name", db_name)
         object.__setattr__(self, "data", pdata)
 
+    def __del__(self):
+        self.api.AdkDeleteStruct(self.data)
+
     def __getattr__(self, key):
-        if key.upper() not in VismaAPI.fields_for_db_field(self.db_name):
-            raise AttributeError("Trying to access field that doesn't exist")
+        try:
+            _type = self.get_type(key)
+        except AttributeError:
+            raise AttributeError(f"{key} is not a valid field of {self.db_name}")
 
-        type = self.get_type(key)
         default_arguments = (self.data, getattr(self.api, key.upper()))
 
-        if type == self.api.ADK_FIELD_TYPE.eChar:
+        if _type == self.api.ADK_FIELD_TYPE.eChar:
             return self.api.AdkGetStr(*default_arguments, String(""))[1]
-        elif type == self.api.ADK_FIELD_TYPE.eDouble:
+        elif _type == self.api.ADK_FIELD_TYPE.eDouble:
             return self.api.AdkGetDouble(*default_arguments, Double(0.0))[1]
-        elif type == self.api.ADK_FIELD_TYPE.eBool:
+        elif _type == self.api.ADK_FIELD_TYPE.eBool:
             return self.api.AdkGetBool(*default_arguments, Boolean(0))[1]
-        elif type == self.api.ADK_FIELD_TYPE.eDate:
+        elif _type == self.api.ADK_FIELD_TYPE.eDate:
             return self.api.AdkGetDate(*default_arguments, DateTime())[1]
 
     def __setattr__(self, key, value):
-        if key.upper() not in VismaAPI.fields_for_db_field(self.db_name):
-            raise AttributeError("Trying to access field that doesn't exist")
+        try:
+            _type = self.get_type(key)
+        except AttributeError:
+            raise AttributeError(f"{key} is not a valid field of {self.db_name}")
 
-        # Make sure user assigns correct type to the field
-        _type = self.get_type(key)
         if not self.assignment_types_are_equal(_type, value):
             raise Exception(f"Trying to assign incorrect type to {key}")
 
@@ -262,7 +268,7 @@ class _Pdata(object):
             error = self.api.AdkSetDate(*default_arguments, self.to_date(value))
 
         if error and error.lRc != self.api.ADKE_OK:
-            error_message = Api.AdkGetErrorText(
+            error_message = self.api.AdkGetErrorText(
                 error, self.api.ADK_ERROR_TEXT_TYPE.elRc
             )
             raise Exception(error_message)
@@ -278,7 +284,7 @@ class _Pdata(object):
         if field_type == self.api.ADK_FIELD_TYPE.eChar and isinstance(input_type, str):
             return True
         elif field_type == self.api.ADK_FIELD_TYPE.eDouble and isinstance(
-            input_type, (float, int)
+            input_type, (float, int, Decimal)
         ):
             return True
         elif field_type == self.api.ADK_FIELD_TYPE.eBool and isinstance(
@@ -312,9 +318,5 @@ class _Pdata(object):
     def delete(self):
         self.api.AdkDeleteRecord(self.data)
 
-
-# Look through fields defined on VismaAPI as ADK_DB_FIELDNAME
-# Dynamically create classes as Fieldname
-for db_field in VismaAPI.db_fields():
-    class_name = VismaAPI.field_without_db_prefix(db_field).title()
-    globals()[class_name] = type(class_name, (_DBField,), {"DB_NAME": db_field})
+    def create(self):
+        self.api.AdkAdd(self.data)
